#define _POSIX_C_SOURCE  200112L
#include <unistd.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <pthread.h>
#include <ctype.h>
#include <signal.h>

#include <conn.h>

static volatile sig_atomic_t termina=0;

static void sighandler(int useless) {
    termina=1;
}

// converte tutti i carattere minuscoli in maiuscoli
void toup(char *str) {
    char *p = str;
    while(*p != '\0') { 
        *p = (islower(*p)?toupper(*p):*p); 
	++p;
    }        
}

void *threadF(void *arg) {
    assert(arg);
    long connfd = (long)arg;
    do {
	msg_t str;
	int n;
	if ((n=readn(connfd, &str.len, sizeof(int))) == -1) {
	    perror("read1");
	    break;
	}

	if (n==0) break;
	str.str = calloc((str.len), sizeof(char));
	if (!str.str) {
	    perror("calloc");
	    fprintf(stderr, "Memoria esaurita....\n");
	    break;
	}		    
	if ((n=readn(connfd, str.str, str.len * sizeof(char))) == -1) {
	    perror("read2");
	    free(str.str);
	    break;
	}

	toup(str.str);

	if ((n=writen(connfd, &str.len, sizeof(int))) == -1) {
	    perror("write1");
	    free(str.str);
	    break;
	}
	if ((n=writen(connfd, str.str, str.len*sizeof(char))) == -1) {
	    perror("write2");
	    free(str.str);
	    break;
	}
	free(str.str);
    } while(1);
    close(connfd);	    
    return NULL;
}

void spawn_thread(long connfd) {
    pthread_attr_t thattr;
    pthread_t thid;

    sigset_t mask,oldmask;
    sigemptyset(&mask);
    sigaddset(&mask, SIGINT); 
    sigaddset(&mask, SIGQUIT);

    if (pthread_sigmask(SIG_BLOCK, &mask,&oldmask) != 0) {
	fprintf(stderr, "FATAL ERROR\n");
	close(connfd);
	return;
    }

    
    if (pthread_attr_init(&thattr) != 0) {
	fprintf(stderr, "pthread_attr_init FALLITA\n");
	close(connfd);
	return;
    }
    // settiamo il thread in modalità detached
    if (pthread_attr_setdetachstate(&thattr,PTHREAD_CREATE_DETACHED) != 0) {
	fprintf(stderr, "pthread_attr_setdetachstate FALLITA\n");
	pthread_attr_destroy(&thattr);
	close(connfd);
	return;
    }
    if (pthread_create(&thid, &thattr, threadF, (void*)connfd) != 0) {
	fprintf(stderr, "pthread_create FALLITA");
	pthread_attr_destroy(&thattr);
	close(connfd);
	return;
    }

    if (pthread_sigmask(SIG_SETMASK, &oldmask, NULL) != 0) {
	fprintf(stderr, "FATAL ERROR\n");
	close(connfd);	
    }
}


int main(int argc, char *argv[]) {
    // installo il signal handler per tutti i segnali che mi interessano
    struct sigaction sa;
    // resetto la struttura
    memset (&sa, 0, sizeof(sa));   
    sa.sa_handler = sighandler;
    int notused;
    SYSCALL(notused,sigaction(SIGINT,   &sa, NULL), "sigaction");
    SYSCALL(notused,sigaction(SIGQUIT,  &sa, NULL), "sigaction");
    SYSCALL(notused,sigaction(SIGTERM,  &sa, NULL), "sigaction");

    int listenfd;
    SYSCALL(listenfd, socket(AF_UNIX, SOCK_STREAM, 0), "socket");

    struct sockaddr_un serv_addr;
    memset(&serv_addr, '0', sizeof(serv_addr));
    serv_addr.sun_family = AF_UNIX;    
    strncpy(serv_addr.sun_path, SOCKNAME, strlen(SOCKNAME)+1);

    SYSCALL(notused, bind(listenfd, (struct sockaddr*)&serv_addr,sizeof(serv_addr)), "bind");
    SYSCALL(notused, listen(listenfd, MAXBACKLOG), "listen");

    while(1) {      
	long connfd;
	if ((connfd= accept(listenfd, (struct sockaddr*)NULL ,NULL)) == -1) {
	    if (errno==EINTR) {
		if (termina) break;
	    } else {
		perror("accept");
		exit(-1);
	    }
	}
	//printf("connection accepted\n");    	
	spawn_thread(connfd);
    }
    // clean-up
    unlink(SOCKNAME);
    return 0;
}
